from .all_courses import AllCoursesQueryData
from .all_courses import DEFINITION as ALL_COURSES_QUERY
from .canvas_course_data import (
    Assignment,
    AssignmentConnection,
    AssignmentGroup,
    AssignmentGroupConnection,
)
from .canvas_course_data import CanvasCourseData as Course
from .canvas_course_data import DEFINITION as COURSE_DATA_FRAGMENT
from .canvas_course_data import (
    File,
    Module,
    ModuleConnection,
    ModuleItem,
    ModuleItemInterface,
    Page,
    Term,
)
from .course_mail import ConversationParticipant, CourseMailQueryData
from .course_mail import DEFINITION as COURSE_MAIL_QUERY
from .single_course import DEFINITION as SINGLE_COURSE_QUERY
from .single_course import SingleCourseQueryData



import qcanvas_backend.database.types as db
from qcanvas_backend.net.sync._internal.canvas_sync_observer import CanvasSyncObserver


class NewContentObserver(CanvasSyncObserver):
    def __init__(self):
        super().__init__()
        self._new_content: list[db.CourseContentItem] = []

    def new_content_found(self, content: object):
        if isinstance(content, db.CourseContentItem):
            self._new_content.append(content)

    @property
    def new_content(self) -> list[db.CourseContentItem]:
        return self._new_content



from pydantic import BaseModel, Field


class IFramePanoptoVideoInfo(BaseModel):
    delivery_id: str = Field(alias="DeliveryId")
    video_url: str = Field(alias="VideoUrl")
    viewer_link: str = Field(alias="ViewerLink")
    thumb_url: str = Field(alias="ThumbUrl")



import logging
from typing import NamedTuple, Optional, Sequence

import bs4
from asynctaskpool import AsyncTaskPool
from asynctaskpool.task_failed_exception import TaskFailedError
from bs4 import Tag

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.extracting.extractors import Extractors
from qcanvas_backend.net.resources.extracting.no_extractor_error import NoExtractorError
from qcanvas_backend.net.resources.scanning.page_resources import PageResources
from qcanvas_backend.task_master import register_reporter
from qcanvas_backend.task_master.reporters import CompoundTaskReporter
from qcanvas_backend.util import CollectingTaskGroup, remove_unwanted_whitespaces
from qcanvas_backend.util.dev_failed_resource_list import FailedResourcesList

_logger = logging.getLogger(__name__)


class _HiddenResource(NamedTuple):
    resource: db.Resource
    is_hidden: bool


class ResourceScanner:
    """
    A ResourceScanner uses Extractors to extract resources from course content
    """

    def __init__(self, extractors: Extractors):
        self._cache = AsyncTaskPool[db.Resource]()
        self._extractor_collection = extractors
        self._failed_resources = FailedResourcesList.create_if_enabled()

    def add_existing_resources(self, existing_resources: Sequence[db.Resource]):
        resources_mapped_by_id = {
            resource.id: resource for resource in existing_resources
        }
        self._cache.update_results(resources_mapped_by_id)

    async def scan_pages_for_resources(
        self, pages: Sequence[db.CourseContentItem]
    ) -> list[PageResources]:
        _logger.info("Scanning %i pages for resources", len(pages))

        if len(pages) == 0:
            return []

        with register_reporter(
            CompoundTaskReporter("Sync", "Scan for resources", len(pages))
        ) as prog:
            async with CollectingTaskGroup() as tg:
                for page in pages:
                    prog.attach(tg.create_task(self._extract_page_resources(page)))

        return tg.results

    async def _extract_page_resources(
        self, page: db.CourseContentItem
    ) -> PageResources:
        async with CollectingTaskGroup() as tg:
            for tag in self._tags_from_page(page):
                tg.create_task(
                    self._extract_resource(
                        tag=tag, course_id=page.course_id, page_id=page.id
                    )
                )

        invisible_resources = []
        visible_resources = []

        for resource, is_invisible in filter(None, tg.results):
            if is_invisible:
                invisible_resources.append(resource)
            else:
                visible_resources.append(resource)

        _logger.debug(
            "Found %i resources (%i invisible) on page %s (id='%s')",
            len(invisible_resources) + len(visible_resources),
            len(invisible_resources),
            page.name,
            page.id,
        )

        return PageResources(
            page_id=page.id,
            resources=visible_resources,
            invisible_resources=invisible_resources,
        )

    async def _extract_resource(
        self, tag: Tag, course_id: str, page_id: str
    ) -> Optional[_HiddenResource]:
        file_id = None

        try:
            extractor = self._extractor_collection.extractor_for_tag(tag)
            file_id = extractor.resource_id_from_tag(tag)

            if await self._failed_resources.failed(file_id):
                _logger.info("File %s has failed before, skipping", file_id)
                return None

            result = await self._cache.submit(
                task_id=file_id,
                future=extractor.resource_from_tag(
                    tag, course_id=course_id, resource_id=file_id
                ),
            )

            return _HiddenResource(
                resource=result, is_hidden=self._is_link_invisible(tag)
            )
        except TaskFailedError as e:
            await self._failed_resources.record_failure(file_id)

            _logger.warning(
                "Extraction failed for file_id=%s on page id=%s",
                file_id or "(no id)",
                page_id,
                exc_info=e,
            )
        except NoExtractorError:
            pass
        except Exception as e:
            _logger.warning("Could not extract resource", exc_info=e)
            pass

    # todo this needs to be a general purpose util function!
    @staticmethod
    def _is_link_invisible(tag: Tag) -> bool:
        return tag.name == "a" and remove_unwanted_whitespaces(tag.text) == ""

    def _tags_from_page(self, page: db.CourseContentItem) -> list[Tag]:
        if page.body is None:
            return []

        doc = bs4.BeautifulSoup(page.body, "html.parser")
        return doc.find_all(self._extractor_collection.tag_whitelist)



import asyncio
import logging
from typing import Union

from lightdb import LightDB
from qcanvas_api_clients.util import enable_api_caching

_logger = logging.getLogger(__name__)


class _NOPFailList:
    async def failed(self, id: str) -> bool:
        return False

    async def record_failure(self, id: str):
        pass


class FailedResourcesList:
    """
    Meant to be used as a development convenience only:
    Remembers if a resource couldn't be retrieved, so we don't bother getting it again. This is saved to disk.
    Should NOT be used for release builds!
    """

    @staticmethod
    def create_if_enabled() -> Union["FailedResourcesList", _NOPFailList]:
        if enable_api_caching:
            _logger.warning(
                'Using development "fail-db". You should not see this message in a release build!'
            )
            return FailedResourcesList()
        else:
            return _NOPFailList()

    def __init__(self):
        self.sem = asyncio.BoundedSemaphore()
        self.db = LightDB("debug_failed_resources.json")

    async def failed(self, id: str) -> bool:
        async with self.sem:
            return self.db.get(id, False)

    async def record_failure(self, id: str):
        async with self.sem:
            self.db[id] = True
            self.db.save()



import logging
from typing import TypeVar

from qcanvas_backend.task_master.task_id import TaskID
from qcanvas_backend.task_master.task_master import Reporter, TaskMaster

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)

T = TypeVar("T", bound=Reporter)


class _DefaultTaskMaster(TaskMaster):

    def report_failed(self, _task_id: TaskID, context: str | Exception) -> None:
        _logger.warning("Task %s failed - %s", context)

    def report_progress(self, _task_id: TaskID, current: int, total: int) -> None:
        if current == total:
            if total == 0:
                _logger.info("Task %s - ?%%", _task_id)
            else:
                _logger.info("Task %s - finished", _task_id)
        else:
            _logger.info(
                "Task %s progress: %s%%", _task_id, f"{(current / total) * 100:.2f}"
            )


_global_task_master = _DefaultTaskMaster()


def get_global_task_master() -> TaskMaster:
    global _global_task_master
    return _global_task_master


def set_global_task_master(new_task_master: TaskMaster) -> None:
    global _global_task_master
    _global_task_master = new_task_master


def register_reporter(reporter: T) -> T:
    return _global_task_master.accept(reporter)



from dataclasses import dataclass
from typing import List

from qcanvas_api_clients.canvas import Announcement

from qcanvas_backend.gql_queries import Course
from qcanvas_backend.net.canvas import CourseMailItem, PageWithContent


@dataclass
class CanvasDataBundle:
    """
    A CanvasDataBundle is a collection of various data retrieved from canvas
    """

    courses: List[Course]
    pages: List[PageWithContent]
    messages: List[CourseMailItem | Announcement]
    course_panopto_folders: dict[str, str]



import logging
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import AsyncIterator, Optional, final

import httpx
from bs4 import Tag

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.download_progress import DownloadProgress

# todo should allow an existing client to be passed into the constructor. we'll see after i figure out how to handle downloading resources
# Use a single client for the generic downloader, making a lot of clients is not a great idea
_generic_download_client = httpx.AsyncClient()

_logger = logging.getLogger(__name__)


class LinkExtractor(ABC):
    """
    Resource ids are formatted using this:
    <extractor name>{RESOURCE_ID_SEPARATOR}<resource id>
    E.g. canvas$12341234
    """

    RESOURCE_ID_SEPARATOR = "$"

    def __init__(
        self,
        extractor_id: str,
        tag_whitelist: list[str],
        is_video_extractor: bool = False,
    ):
        self._extractor_id = extractor_id
        self._tag_whitelist = tag_whitelist
        self._is_video_extractor = is_video_extractor

    @final
    def accepts_tag(self, tag: Tag) -> bool:
        if tag.name not in self.tag_whitelist:
            return False
        else:
            return self._is_tag_valid(tag)

    @abstractmethod
    def _is_tag_valid(self, tag: Tag) -> bool: ...

    @final
    def resource_id_from_tag(self, tag: Tag) -> str:
        return f"{self.extractor_id}{self.RESOURCE_ID_SEPARATOR}{self._tag_id(tag)}"

    @abstractmethod
    def _tag_id(self, tag: Tag) -> str: ...

    @final
    async def resource_from_tag(
        self, tag: Tag, course_id: str, resource_id: str
    ) -> db.Resource:
        resource = await self._extract_resource(tag)
        resource.id = resource_id
        resource.course_id = course_id
        resource.discovery_date = datetime.now()

        return resource

    @abstractmethod
    async def _extract_resource(self, tag: Tag) -> db.Resource: ...

    async def download(
        self, resource: db.Resource, destination: Path
    ) -> AsyncIterator[DownloadProgress]:
        _logger.info("Using generic downloader for %s", resource.id)

        yield DownloadProgress(0, 0)

        finished = False

        async with _generic_download_client.stream(
            url=resource.url, method="get", follow_redirects=True
        ) as response:  # type: httpx.Response
            async for progress in self._download_stream_response(
                destination, response, resource.file_size
            ):
                yield progress

                if progress.downloaded_bytes == progress.total_bytes:
                    finished = True

        if not finished:
            # Make sure we emit a download finished value
            yield DownloadProgress(1, 1)

    @staticmethod
    async def _download_stream_response(
        destination: Path, response: httpx.Response, file_size: Optional[int]
    ) -> AsyncIterator[DownloadProgress]:
        if not response.is_success:
            raise Exception("Download request returned status %i", response.status_code)

        with open(destination, "wb") as out:
            async for chunk in response.aiter_bytes():
                out.write(chunk)

                if file_size is None and "content-length" not in response.headers:
                    _logger.warning(
                        "Content length was missing from response and file size is unknown"
                    )
                    _logger.debug("Headers were %s", ", ".join(response.headers.keys()))

                yield DownloadProgress(
                    # Use 1 as default to prevent division by 0 errors
                    total_bytes=file_size
                    or int(response.headers.get("content-length", 1)),
                    downloaded_bytes=int(response.num_bytes_downloaded),
                )

    @property
    def tag_whitelist(self) -> list[str]:
        return self._tag_whitelist

    @property
    def extractor_id(self) -> str:
        return self._extractor_id

    @property
    def is_video_extractor(self) -> bool:
        return self._is_video_extractor



from bs4 import Tag
from qcanvas_api_clients.canvas import CanvasClient

from qcanvas_backend.database import types as db
from qcanvas_backend.net.resources.extracting.link_extractor import LinkExtractor


class CanvasMediaObjectExtractor(LinkExtractor):
    def __init__(self, canvas_client: CanvasClient):
        super().__init__(
            extractor_id="canvas_media_object",
            tag_whitelist=["iframe"],
            is_video_extractor=True,
        )
        self._canvas_client = canvas_client

    def _is_tag_valid(self, tag: Tag) -> bool:
        return "data-media-type" in tag.attrs and tag["data-media-type"] == "video"

    def _tag_id(self, tag: Tag) -> str:
        return tag["data-media-id"]

    async def _extract_resource(self, tag: Tag) -> db.Resource:
        media_id = tag["data-media-id"]
        media_info = await self._canvas_client.get_media_object(media_id)
        hq_stream = media_info.media_sources[0]  # Highest quality stream is the first

        return db.Resource(
            file_name=media_info.title,
            url=hq_stream.url,
            file_size=int(hq_stream.size) * 1024,  # Given in KiB, not bytes
        )



from pathlib import Path
from typing import Optional, Sequence

from sqlalchemy import select
from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine
from sqlalchemy.orm import selectin_polymorphic

import qcanvas_backend.database.types as db
from qcanvas_backend.database.data_monolith import DataMonolith
from qcanvas_backend.database.types.canvas_database_types import Base


class QCanvasDatabase:
    def __init__(self, database_file: Path):
        self._engine = create_async_engine(f"sqlite+aiosqlite:///{database_file}")
        self._session_maker = async_sessionmaker(self._engine, expire_on_commit=False)

    @staticmethod
    async def create(database_file: Path) -> "QCanvasDatabase":
        new = QCanvasDatabase(database_file)
        await new.init()
        return new

    async def init(self):
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    def session(self):
        # Return type is missing intentionally.
        # Don't try and make this a context manager (which would normally be better) because pycharm still can't
        # determine the type correctly. Blech.
        return self._session_maker.begin()

    async def get_existing_course_ids(self) -> Sequence[str]:
        async with self._session_maker.begin() as session:
            return (await session.scalars(select(db.Course.id))).all()

    async def get_existing_resources(self) -> Sequence[db.Resource]:
        async with self._session_maker.begin() as session:
            stmt = select(db.Resource)
            return (await session.scalars(stmt)).all()

    async def get_resource(self, id: str) -> db.Resource:
        async with self._session_maker.begin() as session:
            stmt = (
                select(db.Resource)
                .where(db.Resource.id == id)
                .options(selectin_polymorphic(db.Resource, [db.PanoptoResource]))
            )
            return (await session.scalars(stmt)).one()

    async def record_resource_downloaded(self, resource: db.Resource):
        await self._update_resource_download_state(
            resource, db.ResourceDownloadState.DOWNLOADED
        )

    async def record_resource_download_failed(
        self, resource: db.Resource, message: str
    ):
        await self._update_resource_download_state(
            resource=resource, message=message, state=db.ResourceDownloadState.FAILED
        )

    async def _update_resource_download_state(
        self,
        resource: db.Resource,
        state: db.ResourceDownloadState,
        message: Optional[str] = None,
    ):
        async with self._session_maker.begin() as session:
            session.add(resource)
            resource.download_state = state
            resource.download_error_message = message

    async def get_data(self) -> DataMonolith:
        async with self.session() as session:
            return await DataMonolith.create(session)



from .canvas_database_types import (
    Assignment,
    AssignmentGroup,
    ContentGroup,
    Course,
    CourseConfiguration,
    CourseContentItem,
    CourseFileGroupingPolicy,
    CourseMessage,
    ModificationDate,
    Module,
    ModulePage,
    ModulePageType,
    PanoptoResource,
    Resource,
    ResourceDownloadState,
    ResourceLink,
    ResourceLinkState,
    Term,
)



from typing import Sequence

from sqlalchemy import select, update
from sqlalchemy.ext.asyncio import AsyncSession

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.scanning.page_resources import PageResources
from qcanvas_backend.net.sync._internal.canvas_sync_observer import CanvasSyncObservable


class ResourceLinker(CanvasSyncObservable):
    def __init__(self, session: AsyncSession):
        super().__init__()
        self._session = session

    async def link_resources(self, page_resources_pairs: list[PageResources]):
        if len(page_resources_pairs) == 0:
            return

        page_ids = [pair.page_id for pair in page_resources_pairs]

        await self._mark_links_as_dead_for_pages(page_ids)
        await self._link_new_resources_and_restore_active_links(page_resources_pairs)

    async def _mark_links_as_dead_for_pages(self, page_ids: list[str]):
        await self._session.execute(
            update(db.ResourceLink)
            .where(db.ResourceLink.content_item_id.in_(page_ids))
            .values(link_state=db.ResourceLinkState.RESIDUAL)
        )

    async def _link_new_resources_and_restore_active_links(
        self, page_resources_pairs: list[PageResources]
    ):
        for pair in page_resources_pairs:
            page_id = pair.page_id
            existing_resource_links = await self._existing_page_resource_links(page_id)
            newly_linked_resources: list[str] = []

            await self._create_links_for_resources(
                existing_resource_links=existing_resource_links,
                page_resource_pair=pair,
                newly_linked_resources=newly_linked_resources,
            )

            await self._create_links_for_invisible_resources(
                existing_resource_links=existing_resource_links,
                page_resource_pair=pair,
                newly_linked_resources=newly_linked_resources,
            )

    async def _create_links_for_resources(
        self,
        existing_resource_links: Sequence[str],
        page_resource_pair: PageResources,
        newly_linked_resources: list[str],
    ):
        page_id = page_resource_pair.page_id

        for resource in page_resource_pair.resources:
            if resource.id in existing_resource_links:
                # If a resource is still linked on this page, reactivate the link
                await self._reactivate_existing_resource_link(
                    page_id=page_id, resource=resource
                )
                continue
            elif resource.id in newly_linked_resources:
                # Prevent adding duplicate links (duplicate links have been observed on some pages)
                continue

            self._create_resource_link(page_id=page_id, resource=resource)
            newly_linked_resources.append(resource.id)

            await self._add_resource_to_db_if_new(resource)

    async def _create_links_for_invisible_resources(
        self,
        existing_resource_links: Sequence[str],
        page_resource_pair: PageResources,
        newly_linked_resources: list[str],
    ):
        page_id = page_resource_pair.page_id

        for resource in page_resource_pair.invisible_resources:
            if (
                resource.id in existing_resource_links
                or resource.id in newly_linked_resources
            ):
                # If the resource was already on the page, or it has just been added, don't add it again
                continue

            self._create_dead_resource_link(page_id=page_id, resource=resource)
            newly_linked_resources.append(resource.id)
            await self._add_resource_to_db_if_new(resource)

    def _create_resource_link(self, page_id: str, resource: db.Resource):
        self._session.add(
            db.ResourceLink(content_item_id=page_id, resource_id=resource.id)
        )

    def _create_dead_resource_link(self, page_id: str, resource: db.Resource):
        link = db.ResourceLink(
            content_item_id=page_id,
            resource_id=resource.id,
            link_state=db.ResourceLinkState.RESIDUAL,
        )
        link.link_state = db.ResourceLinkState.RESIDUAL
        self._session.add(link)

    async def _existing_page_resource_links(self, page_id: str) -> Sequence[str]:
        return (
            await self._session.scalars(
                select(db.ResourceLink.resource_id).where(
                    db.ResourceLink.content_item_id == page_id
                )
            )
        ).all()

    async def _reactivate_existing_resource_link(
        self, page_id: str, resource: db.Resource
    ):
        existing = await self._session.get_one(db.ResourceLink, (page_id, resource.id))
        existing.link_state = db.ResourceLinkState.ACTIVE

    async def _add_resource_to_db_if_new(self, resource: db.Resource):
        does_not_exist = (await self._session.get(db.Resource, resource.id)) is None

        if does_not_exist:
            self._session.add(resource)
            self.notify_observers_for_updated_item(resource)



NBSP = "\N{NO-BREAK SPACE}"


def remove_unwanted_whitespaces(contaminated_data: str | None) -> str | None:
    """
    Remove trailing spaces and replace non-breaking spaces
    :param contaminated_data: The string to clean
    :return: The sanitized string
    """
    if contaminated_data is None:
        return None

    return contaminated_data.replace(NBSP, " ").strip(f"\t {NBSP}")



from typing import Sequence

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import joinedload, selectin_polymorphic, selectinload

import qcanvas_backend.database.types as db


class DataMonolith:
    def __init__(
        self,
        courses: Sequence[db.Course],
        resources: Sequence[db.Resource],
        terms: Sequence[db.Term],
    ):
        self.courses = sorted(courses, key=lambda x: x.name)
        self.resources = self._create_resource_map(resources)
        self.terms = terms

    @staticmethod
    def _create_resource_map(resources):
        return {resource.id: resource for resource in resources}

    @staticmethod
    async def create(session: AsyncSession) -> "DataMonolith":
        eager_load_resources = [
            joinedload(db.Resource.course),
            selectin_polymorphic(db.Resource, [db.PanoptoResource]),
        ]
        eager_load_terms = joinedload(db.Term.courses)
        eager_load_courses = [
            selectinload(db.Course.modules).options(
                joinedload(db.Module.course),
                selectinload(db.Module.pages).joinedload(db.ModulePage.module),
            ),
            selectinload(db.Course.resources).options(*eager_load_resources),
            selectinload(db.Course.content_items).options(
                selectin_polymorphic(
                    db.CourseContentItem,
                    [db.ModulePage, db.CourseMessage, db.Assignment],
                ),
                selectinload(db.CourseContentItem.resources),
                selectinload(db.CourseContentItem.dead_resources),
                joinedload(db.CourseContentItem.course),
            ),
            selectinload(db.Course.messages),
            selectinload(db.Course.assignments),
            selectinload(db.Course.configuration).joinedload(
                db.CourseConfiguration.course
            ),
            selectinload(db.Course.term).joinedload(db.Term.courses),
            selectinload(db.Course.assignment_groups).options(
                joinedload(db.AssignmentGroup.course),
                selectinload(db.AssignmentGroup.assignments).joinedload(
                    db.Assignment.group
                ),
            ),
        ]

        query = select(db.Course).options(*eager_load_courses)
        courses = (await session.scalars(query)).all()
        query = select(db.Resource).options(*eager_load_resources)
        resources = (await session.scalars(query)).all()
        query = select(db.Term).options(eager_load_terms)
        terms = (await session.scalars(query)).unique().all()

        return DataMonolith(courses=courses, resources=resources, terms=terms)



"""
Generated by qenerate plugin=pydantic_v1. DO NOT MODIFY MANUALLY!
"""

from collections.abc import Callable  # noqa: F401 # pylint: disable=W0611
from datetime import datetime  # noqa: F401 # pylint: disable=W0611
from enum import Enum  # noqa: F401 # pylint: disable=W0611
from typing import Any, Optional, Union  # noqa: F401 # pylint: disable=W0611

from pydantic import BaseModel, Extra, Field, Json  # noqa: F401 # pylint: disable=W0611

from qcanvas_backend.gql_queries.canvas_course_data import CanvasCourseData

DEFINITION = """
fragment CanvasCourseData on Course {
    _id
    name
    courseNickname
    term {
        endAt
        startAt
        name
        id
    }
    assignmentGroupsConnection @include(if: $detailed) {
        nodes {
            name
            id
            groupWeight
            assignmentsConnection {
                nodes {
                    description
                    courseId
                    dueAt
                    createdAt
                    id
                    name
                    position
                    updatedAt
                    pointsPossible
                    submissionsConnection {
                        nodes {
                            updatedAt
                            score
                        }
                    }
                }
            }
        }
    }
    modulesConnection @include(if: $detailed) {
        nodes {
            name
            id
            moduleItems {
                content {
                    ... on File {
                        _id
                        displayName
                        createdAt
                        updatedAt
                        url
                        size
                        mimeClass
                        contentType
                    }
                    ... on Page {
                        _id
                        title
                        updatedAt
                        createdAt
                    }
                }
            }
        }
    }
}

query SingleCourse($course_id: ID!, $detailed: Boolean!)
{
    course(id:$course_id) { ... CanvasCourseData }
}
"""


class ConfiguredBaseModel(BaseModel):
    class Config:
        extra = Extra.forbid


class SingleCourseQueryData(ConfiguredBaseModel):
    course: Optional[CanvasCourseData] = Field(..., alias="course")


def query(query_func: Callable, **kwargs: Any) -> SingleCourseQueryData:
    """
    This is a convenience function which queries and parses the data into
    concrete types. It should be compatible with most GQL clients.
    You do not have to use it to consume the generated data classes.
    Alternatively, you can also mime and alternate the behavior
    of this function in the caller.

    Parameters:
        query_func (Callable): Function which queries your GQL Server
        kwargs: optional arguments that will be passed to the query function

    Returns:
        SingleCourseQueryData: queried data parsed into generated classes
    """
    raw_data: dict[Any, Any] = query_func(DEFINITION, **kwargs)
    return SingleCourseQueryData(**raw_data)



import logging
from dataclasses import dataclass, field

import qcanvas_backend.database.types as db
from qcanvas_backend.net.sync._internal.canvas_sync_observer import CanvasSyncObserver

_logger = logging.getLogger(__name__)


@dataclass
class CourseUpdates:
    updated_modules: set[str] = field(default_factory=set)
    updated_pages: set[str] = field(default_factory=set)
    updated_assignment_groups: set[str] = field(default_factory=set)
    updated_assignments: set[str] = field(default_factory=set)
    updated_resources: set[str] = field(default_factory=set)
    updated_messages: set[str | int] = field(default_factory=set)

    def new_content_found(self, content: object) -> None:
        if isinstance(content, db.Module):
            self.updated_modules.add(content.id)
        elif isinstance(content, db.ModulePage):
            self.updated_pages.add(content.id)
            self.updated_modules.add(content.module_id)
        elif isinstance(content, db.CourseMessage):
            self.updated_messages.add(content.id)
        elif isinstance(content, db.AssignmentGroup):
            self.updated_assignment_groups.add(content.id)
        elif isinstance(content, db.Assignment):
            self.updated_assignments.add(content.id)
            self.updated_assignment_groups.add(content.group_id)
        elif isinstance(content, db.Resource):
            self.updated_resources.add(content.id)

    def was_updated(self, content: object) -> bool:
        if isinstance(content, db.Module):
            return content.id in self.updated_modules
        elif isinstance(content, db.ModulePage):
            return content.id in self.updated_pages
        elif isinstance(content, db.CourseMessage):
            return content.id in self.updated_messages
        elif isinstance(content, db.AssignmentGroup):
            return content.id in self.updated_assignment_groups
        elif isinstance(content, db.Assignment):
            return content.id in self.updated_assignments
        elif isinstance(content, db.Resource):
            return content.id in self.updated_resources
        else:
            return False


@dataclass
class SyncReceipt(CanvasSyncObserver, CourseUpdates):
    updated_courses: set[str] = field(default_factory=set)
    updates_by_course: dict[str, CourseUpdates] = field(default_factory=dict)

    def new_content_found(self, content: object) -> None:
        # Need to specify CourseUpdates because of the diamond problem
        CourseUpdates.new_content_found(self, content)

        if isinstance(content, db.Course):
            self.updated_courses.add(content.id)
        else:
            if hasattr(content, "course_id"):
                if content.course_id not in self.updates_by_course:
                    self.updates_by_course[content.course_id] = CourseUpdates()
                    self.updated_courses.add(content.course_id)

                self.updates_by_course[content.course_id].new_content_found(content)

    def was_updated(self, content: object) -> bool:
        if isinstance(content, db.Course):
            return content.id in self.updated_courses
        else:
            return super().was_updated(content)


_empty = SyncReceipt()


def empty_receipt() -> SyncReceipt:
    global _empty
    return _empty



import logging
from typing import *

from qcanvas_backend.task_master.task_master import Reporter

_logger = logging.getLogger(__name__)


class AtomicTaskReporter(Reporter):

    def __init__(self, goal_name: str, step_name: str):
        super().__init__(goal_name, step_name)

    def start(self) -> None:
        self.ensure_task_master_assigned()
        self.task_master.report_progress(self.task_id, 0, 0)

    def finish(self) -> None:
        self.ensure_task_master_assigned()
        self.task_master.report_progress(self.task_id, 1, 1)

    def __enter__(self) -> Self:
        self.start()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if exc_val is not None:
            self.failed(exc_val)
        else:
            self.finish()



"""
Generated by qenerate plugin=pydantic_v1. DO NOT MODIFY MANUALLY!
"""

from collections.abc import Callable  # noqa: F401 # pylint: disable=W0611
from datetime import datetime  # noqa: F401 # pylint: disable=W0611
from enum import Enum  # noqa: F401 # pylint: disable=W0611
from typing import Any, Optional, Union  # noqa: F401 # pylint: disable=W0611

from pydantic import BaseModel, Extra, Field, Json  # noqa: F401 # pylint: disable=W0611

DEFINITION = """
fragment CanvasCourseData on Course {
    _id
    name
    courseNickname
    term {
        endAt
        startAt
        name
        id
    }
    assignmentGroupsConnection @include(if: $detailed) {
        nodes {
            name
            id
            groupWeight
            assignmentsConnection {
                nodes {
                    description
                    courseId
                    dueAt
                    createdAt
                    id
                    name
                    position
                    updatedAt
                    pointsPossible
                    submissionsConnection {
                        nodes {
                            updatedAt
                            score
                        }
                    }
                }
            }
        }
    }
    modulesConnection @include(if: $detailed) {
        nodes {
            name
            id
            moduleItems {
                content {
                    ... on File {
                        _id
                        displayName
                        createdAt
                        updatedAt
                        url
                        size
                        mimeClass
                        contentType
                    }
                    ... on Page {
                        _id
                        title
                        updatedAt
                        createdAt
                    }
                }
            }
        }
    }
}
"""


class ConfiguredBaseModel(BaseModel):
    class Config:
        extra = Extra.forbid


class Term(ConfiguredBaseModel):
    end_at: Optional[datetime] = Field(..., alias="endAt")
    start_at: Optional[datetime] = Field(..., alias="startAt")
    name: Optional[str] = Field(..., alias="name")
    q_id: str = Field(..., alias="id")


class Submission(ConfiguredBaseModel):
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")
    score: Optional[float] = Field(..., alias="score")


class SubmissionConnection(ConfiguredBaseModel):
    nodes: Optional[list[Optional[Submission]]] = Field(..., alias="nodes")


class Assignment(ConfiguredBaseModel):
    description: Optional[str] = Field(..., alias="description")
    course_id: Optional[str] = Field(..., alias="courseId")
    due_at: Optional[datetime] = Field(..., alias="dueAt")
    created_at: Optional[datetime] = Field(..., alias="createdAt")
    q_id: str = Field(..., alias="id")
    name: Optional[str] = Field(..., alias="name")
    position: Optional[int] = Field(..., alias="position")
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")
    points_possible: Optional[float] = Field(..., alias="pointsPossible")
    submissions_connection: Optional[SubmissionConnection] = Field(
        ..., alias="submissionsConnection"
    )


class AssignmentConnection(ConfiguredBaseModel):
    nodes: Optional[list[Optional[Assignment]]] = Field(..., alias="nodes")


class AssignmentGroup(ConfiguredBaseModel):
    name: Optional[str] = Field(..., alias="name")
    q_id: str = Field(..., alias="id")
    group_weight: Optional[float] = Field(..., alias="groupWeight")
    assignments_connection: Optional[AssignmentConnection] = Field(
        ..., alias="assignmentsConnection"
    )


class AssignmentGroupConnection(ConfiguredBaseModel):
    nodes: Optional[list[Optional[AssignmentGroup]]] = Field(..., alias="nodes")


class ModuleItemInterface(ConfiguredBaseModel): ...


class File(ModuleItemInterface):
    q_id: str = Field(..., alias="_id")
    display_name: Optional[str] = Field(..., alias="displayName")
    created_at: Optional[datetime] = Field(..., alias="createdAt")
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")
    url: Optional[str] = Field(..., alias="url")
    size: Optional[str] = Field(..., alias="size")
    mime_class: Optional[str] = Field(..., alias="mimeClass")
    content_type: Optional[str] = Field(..., alias="contentType")


class Page(ModuleItemInterface):
    q_id: str = Field(..., alias="_id")
    title: Optional[str] = Field(..., alias="title")
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")
    created_at: Optional[datetime] = Field(..., alias="createdAt")


class ModuleItem(ConfiguredBaseModel):
    content: Optional[Union[File, Page, ModuleItemInterface]] = Field(
        ..., alias="content"
    )


class Module(ConfiguredBaseModel):
    name: Optional[str] = Field(..., alias="name")
    q_id: str = Field(..., alias="id")
    module_items: Optional[list[ModuleItem]] = Field(..., alias="moduleItems")


class ModuleConnection(ConfiguredBaseModel):
    nodes: Optional[list[Optional[Module]]] = Field(..., alias="nodes")


class CanvasCourseData(ConfiguredBaseModel):
    q_id: str = Field(..., alias="_id")
    name: str = Field(..., alias="name")
    course_nickname: Optional[str] = Field(..., alias="courseNickname")
    term: Optional[Term] = Field(..., alias="term")
    assignment_groups_connection: Optional[AssignmentGroupConnection] = Field(
        None, alias="assignmentGroupsConnection"
    )
    modules_connection: Optional[ModuleConnection] = Field(
        None, alias="modulesConnection"
    )



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from qcanvas_backend.gql_queries import Course, Module


@dataclass
class PageWithContent:
    q_id: str
    name: Optional[str]
    updated_at: Optional[datetime]
    created_at: Optional[datetime]
    module: Module
    course: Course
    position: int
    content: Optional[str] = None



import logging
from asyncio import Task
from threading import Lock
from typing import Self

from qcanvas_backend.task_master.task_master import Reporter

_logger = logging.getLogger(__name__)


class CompoundTaskReporter(Reporter):
    def __init__(self, goal_name: str, step_name: str, total_tasks: int):
        super().__init__(goal_name, step_name)
        self._lock = Lock()
        self._total_tasks = total_tasks
        self._complete_tasks = 0

    def __enter__(self) -> Self:
        self.ensure_task_master_assigned()
        self.progress(0, self._total_tasks)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if exc_val is not None:
            self.failed(exc_val)

    def attach(self, task: Task) -> Task:
        task.add_done_callback(self._increment_progress)
        return task

    def _increment_progress(self, _) -> None:
        # Contention for this should be extremely low, so it shouldn't have much impact on the event loop
        with self._lock:
            self._complete_tasks += 1
            progress = self._complete_tasks

        self.progress(progress, self._total_tasks)



import logging
import platform
import re
from abc import ABC, abstractmethod
from asyncio import BoundedSemaphore, TaskGroup
from pathlib import Path, PurePath
from typing import *

from asynctaskpool import AsyncTaskPool

import qcanvas_backend.database.types as db
from qcanvas_backend.database import QCanvasDatabase
from qcanvas_backend.net.resources.extracting.extractors import Extractors
from qcanvas_backend.net.resources.extracting.link_extractor import LinkExtractor
from qcanvas_backend.net.resources.scanning.resource_scanner import ResourceScanner
from qcanvas_backend.task_master import register_reporter
from qcanvas_backend.task_master.reporters import ContextManagerReporter

_logger = logging.getLogger(__name__)


class ResourceManager(ABC):
    def __init__(
        self,
        database: QCanvasDatabase,
        download_dest: Path,
        extractors: Extractors = Extractors(),
        max_concurrent_downloads: int = 8,
    ):
        super().__init__()
        self._download_pool = AsyncTaskPool[None]()
        self._db = database
        self._extractors = extractors
        self._scanner = ResourceScanner(extractors)
        self._download_dest = download_dest
        self._download_limiter = BoundedSemaphore(max_concurrent_downloads)

    def add_existing_resources(self, existing_resources: Sequence[db.Resource]):
        self._scanner.add_existing_resources(existing_resources)

    def is_video(self, resource: db.Resource) -> bool:
        # todo using the extractor to tell if its a video is kinda lacklustre
        return self._extractors.extractor_for_resource(resource).is_video_extractor

    async def batch_download(
        self,
        resources: list[db.Resource],
    ) -> None:
        downloads: List[Tuple[LinkExtractor, db.Resource]] = []
        deferred_downloads: List[Tuple[LinkExtractor, db.Resource]] = []

        for resource in resources:
            extractor = self._extractors.extractor_for_resource(resource)

            # Defer large downloads to do them 1 at a time
            if extractor.is_video_extractor:
                deferred_downloads.append((extractor, resource))
            else:
                downloads.append((extractor, resource))

        async with TaskGroup() as tg:
            for extractor, resource in downloads:
                tg.create_task(
                    self._start_download(extractor, resource, suppress_exceptions=True)
                )

        # Do large downloads one at a time. Parallelizing them is unlikely to speed anything up.
        for extractor, resource in deferred_downloads:
            await self._start_download(extractor, resource, suppress_exceptions=True)

    async def download(self, resource: db.Resource) -> None:
        if resource.download_state != db.ResourceDownloadState.DOWNLOADED:
            extractor = self._extractors.extractor_for_resource(resource)
            await self._start_download(extractor, resource)

    async def _start_download(
        self,
        extractor: LinkExtractor,
        resource: db.Resource,
        suppress_exceptions: bool = False,
    ) -> None:
        await self._download_pool.submit(
            resource.id,
            self._download_task(
                extractor, resource, suppress_exceptions=suppress_exceptions
            ),
        )

    async def _download_task(
        self,
        extractor: LinkExtractor,
        resource: db.Resource,
        *,
        suppress_exceptions: bool = False,
    ) -> None:
        async with self._download_limiter:
            try:
                destination = self.resource_download_location(resource)
                destination.parent.mkdir(parents=True, exist_ok=True)

                with register_reporter(
                    ContextManagerReporter("Download", resource.file_name, 0)
                ) as prog:  # type: ContextManagerReporter
                    async for progress in extractor.download(
                        resource,
                        destination=destination,
                    ):
                        prog.progress(progress.downloaded_bytes, progress.total_bytes)
                        self.on_download_progress(
                            resource=resource,
                            current=progress.downloaded_bytes,
                            total=progress.total_bytes,
                        )
            except Exception as e:
                await self._db.record_resource_download_failed(resource, message=str(e))
                self.on_download_failed(resource)

                if not suppress_exceptions:
                    raise RuntimeError() from e
                else:
                    _logger.warning("Download failed", exc_info=e)
            else:
                await self._db.record_resource_downloaded(resource)
                self.on_download_finished(resource)

    def resource_download_location(self, resource: db.Resource) -> Path:
        file_name, file_suffix = ResourceManager._split_resource_name(resource)
        file_suffix = f" [{resource.id}]{file_suffix}"
        # Ensure the filename is not too long, most filesystems permit up to 255 chars
        file_name = file_name[: 255 - len(file_suffix)]

        return (
            self._download_dest
            / self._replace_illegal_chars(resource.course.name)
            / self._replace_illegal_chars(file_name + file_suffix)
        )

    @staticmethod
    def _split_resource_name(resource: db.Resource) -> Tuple[str, str]:
        file = PurePath(resource.file_name)
        return file.stem, file.suffix

    @staticmethod
    def _replace_illegal_chars(file_name: str) -> str:
        if platform.system() == "Windows":
            return re.sub(r"[<>:\"/\\|?*]", "_", file_name)
        else:
            return file_name.replace("/", "_")

    @property
    def scanner(self) -> ResourceScanner:
        return self._scanner

    @property
    def extractors(self) -> Extractors:
        return self._extractors

    @property
    def downloads_folder(self) -> Path:
        return self._download_dest

    @abstractmethod
    def on_download_progress(
        self, resource: db.Resource, current: int, total: int
    ) -> None: ...

    @abstractmethod
    def on_download_failed(self, resource: db.Resource) -> None: ...

    @abstractmethod
    def on_download_finished(self, resource: db.Resource) -> None: ...



import functools
import operator
from typing import Iterable, TypeVar

_T = TypeVar("_T")
_V = TypeVar("_V")


def flatten(the_list: Iterable[Iterable[_T]]) -> list[_T]:
    return functools.reduce(operator.iconcat, the_list, [])



import asyncio
import logging
from pathlib import Path

from qcanvas_api_clients.canvas import Announcement, CanvasClient
from qcanvas_api_clients.panopto import PanoptoClient

import qcanvas_backend.gql_queries as gql
from qcanvas_backend.database import QCanvasDatabase
from qcanvas_backend.net.canvas import (
    CanvasDataAggregator,
    CanvasPageLoader,
    CourseMailItem,
)
from qcanvas_backend.net.panopto.panopto_data_aggregator import (
    PanoptoDataAggregator,
)
from qcanvas_backend.net.resources.scanning.resource_linker import ResourceLinker
from qcanvas_backend.net.resources.scanning.resource_scanner import ResourceScanner
from qcanvas_backend.net.sync._internal.canvas_data_bundle import CanvasDataBundle
from qcanvas_backend.net.sync._internal.new_content_observer import NewContentObserver
from qcanvas_backend.net.sync._internal.sync_meta import SyncMeta
from qcanvas_backend.net.sync.api_data_importer import APIDataImporter
from qcanvas_backend.net.sync.data_adapter import DataAdapter
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qcanvas_backend.task_master import register_reporter
from qcanvas_backend.task_master.reporters import (
    AtomicTaskReporter,
    CompoundTaskReporter,
)
from qcanvas_backend.util import CollectingTaskGroup, flatten

_logger = logging.getLogger(__name__)


class SynchronizationClient:
    """
    A SynchronizationClient uses a CanvasClient to retrieve data from canvas and a ResourceScanner to extract resources
    from each course. This data is added to the QCanvasDatabase.
    """

    def __init__(
        self,
        database: QCanvasDatabase,
        canvas_client: CanvasClient,
        panopto_client: PanoptoClient,
        resource_scanner: ResourceScanner,
        sync_meta_store_location: Path,
    ):
        self._database = database
        self._canvas_client = canvas_client
        self._canvas_data_aggregator = CanvasDataAggregator(
            canvas_client=self._canvas_client
        )
        self._panopto_data_aggregator = PanoptoDataAggregator(
            panopto_client=panopto_client
        )
        self._sync_meta = SyncMeta(sync_meta_store_location)
        self._canvas_page_loader = CanvasPageLoader(canvas_client=self._canvas_client)
        self._resource_scanner = resource_scanner

    async def synchronise_canvas(
        self, include_old_courses: bool = False
    ) -> SyncReceipt:
        receipt = SyncReceipt()
        new_content_tracker = NewContentObserver()
        sync_observers = [receipt, new_content_tracker]

        user_id = await self._canvas_client.get_current_user_id()
        canvas_sync_data = await self._fetch_remote_sync_data(
            user_id, include_old_courses=include_old_courses
        )

        async with self._database.session() as session:
            importer = APIDataImporter(session)
            resource_linker = ResourceLinker(session)
            adapter = DataAdapter(importer)

            importer.observers.extend(sync_observers)
            resource_linker.observers.extend(sync_observers)

            await adapter.convert_and_add_to_database(canvas_sync_data)

            scanned_resources = await self._resource_scanner.scan_pages_for_resources(
                new_content_tracker.new_content
            )
            await resource_linker.link_resources(scanned_resources)

        self._sync_meta.update_last_sync_time()

        _logger.info("Sync finished")
        _logger.debug(receipt)

        return receipt

    async def _fetch_remote_sync_data(
        self, user_id: str, include_old_courses: bool
    ) -> CanvasDataBundle:
        _logger.info("Last sync time is %s", self._sync_meta.last_sync_time)

        courses = await self._canvas_data_aggregator.pull_courses(
            await self._database.get_existing_course_ids(),
            include_old_courses=not include_old_courses,
        )
        pages = await self._canvas_page_loader.load_all_updated_page_content(
            courses=courses, last_update_time=self._sync_meta.last_sync_time
        )

        mail: list[CourseMailItem] = []
        announcements: list[Announcement] = []
        course_panopto_folders: dict[str, str] = {}

        async def get_mail() -> None:
            nonlocal mail
            nonlocal user_id
            mail = await self._canvas_data_aggregator.get_all_course_mail(user_id)

        async def get_announcements() -> None:
            nonlocal announcements
            nonlocal courses
            announcements = flatten(await self._get_announcements_for_courses(courses))

        # async def get_panopto_folders():
        #     nonlocal course_panopto_folders
        #     nonlocal courses
        #     course_panopto_folders = (
        #         await self._panopto_data_aggregator.get_course_panopto_folders(courses)
        #     )

        async def dummy_panopto_task() -> None:
            """
            Run a panopto request so that the client authenticates itself.
            This should prevent the resource scan from pausing partway through (and confusing the user) due to needing to authenticate.
            """
            with register_reporter(
                AtomicTaskReporter("sync", "Panopto authentication")
            ):
                await self._panopto_data_aggregator.get_folders()

        await asyncio.gather(get_mail(), get_announcements(), dummy_panopto_task())

        _logger.info(
            "%i courses, %i pages, %i mails, and %i announcements pulled from canvas",
            len(courses),
            len(pages),
            len(mail),
            len(announcements),
        )

        return CanvasDataBundle(
            courses=courses,
            pages=pages,
            messages=mail + announcements,
            course_panopto_folders=course_panopto_folders,
        )

    async def _get_announcements_for_courses(
        self, courses: list[gql.Course]
    ) -> list[Announcement]:
        _logger.info("Fetching announcements for %i courses", len(courses))

        if len(courses) == 0:
            return []

        with register_reporter(
            CompoundTaskReporter("Sync", "Fetch announcements", len(courses))
        ) as prog:
            async with CollectingTaskGroup() as tg:
                for course in courses:
                    prog.attach(
                        tg.create_task(
                            self._canvas_client.get_announcements(course.q_id)
                        )
                    )

        return tg.results



import logging
from typing import Sequence

from gql import gql
from qcanvas_api_clients.canvas import CanvasClient
from qcanvas_api_clients.util import cachedasyncmethod_for_development
from shelved_cache.keys import autotuple_hashkey

from qcanvas_backend import gql_queries
from qcanvas_backend.gql_queries import (
    AllCoursesQueryData,
    ConversationParticipant,
    Course,
    CourseMailQueryData,
    Term,
)
from qcanvas_backend.net.canvas.course_mail_item import CourseMailItem
from qcanvas_backend.task_master import register_reporter
from qcanvas_backend.task_master.reporters import (
    AtomicTaskReporter,
    CompoundTaskReporter,
)
from qcanvas_backend.util import CollectingTaskGroup

_logger = logging.getLogger(__name__)


class CanvasDataAggregator:
    def __init__(self, canvas_client: CanvasClient):
        self._canvas_client = canvas_client

    async def pull_user_id(self) -> str:
        return await self._canvas_client.get_current_user_id()

    @cachedasyncmethod_for_development(key=autotuple_hashkey)
    async def get_all_course_mail(self, current_user_id: str) -> list[CourseMailItem]:
        _logger.info("Fetching all course mail")

        with register_reporter(AtomicTaskReporter("Sync", "Fetch course mail")):
            result = await self._gql_query(
                gql_queries.COURSE_MAIL_QUERY, user_id=current_user_id
            )

        raw_mail_data = CourseMailQueryData(
            **result
        ).legacy_node.conversations_connection.nodes

        return self._convert_mail(raw_mail_data)

    @staticmethod
    def _convert_mail(
        raw_mail_data: list[ConversationParticipant],
    ):
        return [CourseMailItem.from_query_result(mail) for mail in raw_mail_data]

    @cachedasyncmethod_for_development(key=autotuple_hashkey)
    async def pull_courses(
        self, already_indexed_course_ids: Sequence[str], include_old_courses: bool
    ) -> list[Course]:
        shallow_courses = await self._get_shallow_course_data()
        shallow_courses = self._ignore_blackboard_courses(shallow_courses)
        latest_term = self._find_latest_term(shallow_courses)

        course_ids_to_pull = [
            course.q_id
            for course in shallow_courses
            if self._course_belongs_to_term(course, latest_term)
            or (include_old_courses and course.q_id not in already_indexed_course_ids)
        ]

        return await self._pull_detailed_courses_by_id(course_ids_to_pull)

    @staticmethod
    def _course_belongs_to_term(course: Course, term: Term) -> bool:
        return course.term.q_id == term.q_id

    async def _get_shallow_course_data(self) -> list[Course]:
        result = await self._gql_query(gql_queries.ALL_COURSES_QUERY, detailed=False)

        return AllCoursesQueryData(**result).all_courses

    @staticmethod
    def _ignore_blackboard_courses(course_list: list[Course]) -> list[Course]:
        return [
            course
            for course in course_list
            if course.term.start_at is not None and course.term.start_at.year >= 2022
        ]

    async def _pull_detailed_courses_by_id(self, course_ids: list[str]) -> list[Course]:
        _logger.info("Fetching index for course ids %s", ", ".join(course_ids))

        if len(course_ids) == 0:
            return []

        with register_reporter(
            CompoundTaskReporter("Sync", "Fetch course data", len(course_ids))
        ) as prog:
            async with CollectingTaskGroup() as tg:
                for course_id in course_ids:
                    prog.attach(
                        tg.create_task(
                            self._gql_query(
                                gql_queries.SINGLE_COURSE_QUERY,
                                course_id=course_id,
                                detailed=True,
                            )
                        )
                    )

        return self._convert_courses(tg.results)

    async def _gql_query(self, query: str, **kwargs):
        return await self._canvas_client.graphql_query(gql(query), kwargs)

    @staticmethod
    def _convert_courses(results: list[dict]) -> list[Course]:
        return [
            gql_queries.SingleCourseQueryData(**task_result).course
            for task_result in results
        ]

    @staticmethod
    def _find_latest_term(courses: list[Course]) -> Term:
        terms = [course.term for course in courses if course.term is not None]
        sorted_terms = sorted(terms, key=lambda term: term.end_at)

        return sorted_terms[-1]



import logging
from datetime import datetime
from pathlib import Path

from lightdb import LightDB

_logger = logging.getLogger(__name__)


class SyncMeta:
    def __init__(self, db_path: Path):
        self._db = LightDB(str(db_path.absolute()))

    @property
    def last_sync_time(self) -> datetime:
        if "last" not in self._db:
            return datetime.min
        else:
            return datetime.fromisoformat(self._db["last"])

    def update_last_sync_time(self):
        self._db["last"] = datetime.now().isoformat()
        self._db.save()



import logging
from asyncio import TaskGroup
from datetime import datetime

from qcanvas_api_clients.canvas import CanvasClient

from qcanvas_backend.gql_queries import Course, Module, ModuleItem, Page
from qcanvas_backend.net.canvas.page_with_content import PageWithContent
from qcanvas_backend.task_master import register_reporter
from qcanvas_backend.task_master.reporters import CompoundTaskReporter
from qcanvas_backend.util import strip_tz

_logger = logging.getLogger(__name__)


class CanvasPageLoader:

    def __init__(self, canvas_client: CanvasClient):
        self._canvas_client = canvas_client

    async def load_all_updated_page_content(
        self, courses: list[Course], last_update_time: datetime
    ) -> list[PageWithContent]:
        flattened_page_list = self._get_outdated_page_list(
            courses=courses, last_update_time=last_update_time
        )

        _logger.info("Loading %i pages", len(flattened_page_list))

        if len(flattened_page_list) == 0:
            return []

        with register_reporter(
            CompoundTaskReporter("Sync", "Load pages", len(flattened_page_list))
        ) as prog:
            async with TaskGroup() as tg:
                for page in flattened_page_list:
                    _logger.debug(
                        '%s "%s" (id="%s") out of date',
                        type(page).__name__,
                        page.name,
                        page.q_id,
                    )
                    prog.attach(tg.create_task(self._load_page_content_inplace(page)))

        return flattened_page_list

    def _get_outdated_page_list(
        self, courses: list[Course], last_update_time: datetime
    ) -> list[PageWithContent]:
        pages = []

        for course in courses:
            for module in course.modules_connection.nodes:
                for position, module_item in enumerate(module.module_items):
                    if isinstance(module_item.content, Page):
                        if self._is_module_item_out_of_date(
                            last_update_time, module_item
                        ):
                            pages.append(
                                self._convert_to_page_with_content(
                                    page=module_item.content,
                                    module=module,
                                    course=course,
                                    position=position,
                                )
                            )

        return pages

    @staticmethod
    def _is_module_item_out_of_date(
        last_update_time: datetime, module_item: ModuleItem
    ) -> bool:
        return strip_tz(module_item.content.updated_at) >= last_update_time

    @staticmethod
    def _convert_to_page_with_content(
        page: Page, module: Module, course: Course, position: int
    ):
        return PageWithContent(
            q_id=page.q_id,
            name=page.title,
            updated_at=page.updated_at,
            created_at=page.created_at,
            module=module,
            course=course,
            position=position,
        )

    async def _load_page_content_inplace(self, page: PageWithContent):
        result = await self._canvas_client.get_page(
            page_id=page.q_id, course_id=page.course.q_id
        )
        _logger.debug('Loaded page %s (id="%s")', page.name, page.q_id)
        page.content = result.body



import logging

from qcanvas_api_clients.canvas import Announcement

import qcanvas_backend.gql_queries as gql
from qcanvas_backend.net.canvas import CourseMailItem, PageWithContent
from qcanvas_backend.net.sync._internal.canvas_data_bundle import CanvasDataBundle
from qcanvas_backend.net.sync.api_data_importer import APIDataImporter
from qcanvas_backend.task_master import register_reporter
from qcanvas_backend.task_master.reporters import AtomicTaskReporter

_logger = logging.getLogger(__name__)


class DataAdapter:
    """
    A DataAdapter deeply iterates over everything in a CanvasDataBundle and adds this data to the database using an APIDataImporter
    """

    def __init__(self, importer: APIDataImporter):
        self._converter = importer

    async def convert_and_add_to_database(self, canvas_sync_data: CanvasDataBundle):
        with register_reporter(AtomicTaskReporter("Sync", "Add data to database")):
            await self._store_partial_course_data(
                courses=canvas_sync_data.courses,
                course_panopto_folders=canvas_sync_data.course_panopto_folders,
            )
            await self._store_pages(canvas_sync_data.pages)
            await self._store_messages(
                canvas_sync_data.messages,
                known_course_ids=[course.q_id for course in canvas_sync_data.courses],
            )

    async def _store_partial_course_data(
        self, courses: list[gql.Course], course_panopto_folders: dict[str, str]
    ):
        for term in self._flatten_terms(courses):
            await self._converter.convert_and_store_term(term)

        for course in courses:
            await self._converter.convert_and_store_course(
                course=course,
                panopto_folder_id=course_panopto_folders.get(course.q_id, None),
            )

            for position, module in enumerate(course.modules_connection.nodes):
                await self._converter.convert_and_store_module(
                    module=module, course_id=course.q_id, position=position
                )

            for position, assignment_group in enumerate(
                course.assignment_groups_connection.nodes
            ):
                if self._assignment_group_is_empty(assignment_group):
                    continue

                await self._converter.convert_and_store_assignment_group(
                    assignment_group=assignment_group,
                    course_id=course.q_id,
                    position=position,
                )

                for assignment in assignment_group.assignments_connection.nodes:
                    await self._converter.convert_and_store_assignment(
                        assignment=assignment, group_id=assignment_group.q_id
                    )

    @staticmethod
    def _assignment_group_is_empty(assignment_group: gql.AssignmentGroup) -> bool:
        return len(assignment_group.assignments_connection.nodes) == 0

    @staticmethod
    def _flatten_terms(courses: list[gql.Course]) -> list[gql.Term]:
        term_id_map: dict[str, gql.Term] = {}

        for course in courses:
            term_id_map[course.term.q_id] = course.term

        return list(term_id_map.values())

    async def _store_pages(self, pages: list[PageWithContent]):
        for page in pages:
            await self._converter.convert_and_store_page(page)

    async def _store_messages(
        self, messages: list[CourseMailItem | Announcement], known_course_ids: list[str]
    ):
        for message in messages:
            if isinstance(message, CourseMailItem):
                # Canvas includes mail for information "courses", but they're not visible in the allCourses graphql item.
                # Since they are not indexed anywhere else, we will discard them here
                if message.course_id in known_course_ids:
                    await self._converter.convert_and_store_mail_item(message)
                else:
                    _logger.debug(
                        'Discarding mail item "%s" (id="%s") because it\'s from an un-indexed course (id="%s")',
                        message.subject,
                        message.id,
                        message.course_id,
                    )
            elif isinstance(message, Announcement):
                await self._converter.convert_and_store_announcement(message)



from datetime import datetime
from typing import List, Optional, Protocol, Sequence

from sqlalchemy import ForeignKey
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    MappedAsDataclass,
    mapped_column,
    relationship,
)

from qcanvas_backend.database.types.course_page_grouping_policy import (
    CourseFileGroupingPolicy,
)
from qcanvas_backend.database.types.module_page_type import ModulePageType
from qcanvas_backend.database.types.resource_download_state import ResourceDownloadState
from qcanvas_backend.database.types.resource_life_state import ResourceLinkState


class ModificationDate:
    last_modification_date: Mapped[datetime]


class Base(DeclarativeBase, MappedAsDataclass, init=False):
    pass


class ResourceLink(Base, init=True):
    __tablename__ = "resource_links"

    # Do not reorder these (it will mess up the key tuple)!! content_item_id, then resource_id
    content_item_id: Mapped[str] = mapped_column(
        ForeignKey("course_content.id"), primary_key=True
    )
    resource_id: Mapped[str] = mapped_column(
        ForeignKey("resources.id"), primary_key=True
    )
    link_state: Mapped[ResourceLinkState] = mapped_column(
        default=ResourceLinkState.ACTIVE
    )


class Resource(Base):
    __tablename__ = "resources"

    id: Mapped[str] = mapped_column(primary_key=True)

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="resources", cascade="merge")

    url: Mapped[str]
    file_name: Mapped[str]
    file_size: Mapped[Optional[int]] = mapped_column(default=None)
    discovery_date: Mapped[datetime]
    download_state: Mapped[ResourceDownloadState] = mapped_column(
        default=ResourceDownloadState.NOT_DOWNLOADED
    )
    download_error_message: Mapped[Optional[str]] = mapped_column(default=None)

    polymorphic_type: Mapped[str]

    __mapper_args__ = {
        "polymorphic_on": "polymorphic_type",
        "polymorphic_identity": "resource",
    }


class CourseContentItem(Base):
    __tablename__ = "course_content"

    id: Mapped[str] = mapped_column(primary_key=True)

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="content_items")

    name: Mapped[str]
    body: Mapped[Optional[str]]
    creation_date: Mapped[datetime]
    polymorphic_type: Mapped[str]

    resources: Mapped[List["Resource"]] = relationship(
        secondary=ResourceLink.__table__,
        primaryjoin=f"and_(ResourceLink.content_item_id == CourseContentItem.id, ResourceLink.link_state=='{ResourceLinkState.ACTIVE.name}')",
        overlaps="dead_resources, course_content",
        order_by=Resource.discovery_date,
    )
    dead_resources: Mapped[List["Resource"]] = relationship(
        secondary=ResourceLink.__table__,
        primaryjoin=f"and_(ResourceLink.content_item_id == CourseContentItem.id, ResourceLink.link_state=='{ResourceLinkState.RESIDUAL.name}')",
        overlaps="resources, course_content",
        order_by=Resource.discovery_date,
    )

    __mapper_args__ = {
        "polymorphic_on": "polymorphic_type",
        "polymorphic_identity": "generic_content",
    }


class ContentGroup(Protocol):
    @property
    def id(self) -> str: ...

    @property
    def name(self) -> str: ...

    @property
    def content_items(self) -> Sequence[CourseContentItem]: ...


class CourseConfiguration(Base):
    __tablename__ = "course_configuration"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(
        back_populates="configuration", single_parent=True
    )

    nickname: Mapped[Optional[str]] = mapped_column(default=None)
    file_grouping_policy: Mapped[CourseFileGroupingPolicy] = mapped_column(
        default=CourseFileGroupingPolicy.GROUP_BY_PAGES
    )


class PanoptoResource(Resource):
    __tablename__ = "panopto_resources"

    id: Mapped[str] = mapped_column(ForeignKey("resources.id"), primary_key=True)

    duration_seconds: Mapped[int]
    recording_date: Mapped[datetime]

    # Panopto/Canvas have this stupid "custom_context_delivery" which is a pain in the ass because it has nothing to do
    # with the actual ID of the video. In this case, this object's id may not be useful in any way (thanks painopto),
    # but delivery_id will always be the true ID of the video.
    # !!!! It has been observed that different "custom_context_delivery"s CAN link to the same video !!!!
    delivery_id: Mapped[str]

    __mapper_args__ = {"polymorphic_identity": "panopto_resource"}


class Assignment(CourseContentItem, ModificationDate):
    __tablename__ = "assignments"

    id: Mapped[str] = mapped_column(ForeignKey("course_content.id"), primary_key=True)

    due_date: Mapped[Optional[datetime]]
    mark: Mapped[Optional[float]]
    max_mark_possible: Mapped[Optional[float]]
    position: Mapped[int]

    group_id: Mapped[str] = mapped_column(ForeignKey("assignment_groups.id"))
    group: Mapped["AssignmentGroup"] = relationship(back_populates="assignments")

    __mapper_args__ = {"polymorphic_identity": "assignment"}


class AssignmentGroup(Base):
    __tablename__ = "assignment_groups"

    id: Mapped[str] = mapped_column(primary_key=True)

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="assignment_groups")

    name: Mapped[str]
    assignments: Mapped[List["Assignment"]] = relationship(
        back_populates="group",
        order_by=Assignment.position,
        cascade="save-update, merge, delete",
    )
    group_weight: Mapped[int]
    position: Mapped[int]

    @property
    def content_items(self) -> Sequence[CourseContentItem]:
        """
        Alias for .assignments
        """
        return self.assignments


class CourseMessage(CourseContentItem):
    """
    Used for announcements and course mail
    """

    __tablename__ = "messages"

    id: Mapped[str] = mapped_column(ForeignKey("course_content.id"), primary_key=True)

    sender_name: Mapped[str]
    has_been_read: Mapped[bool]

    __mapper_args__ = {"polymorphic_identity": "message"}


class ModulePage(CourseContentItem, ModificationDate):
    __tablename__ = "pages"

    id: Mapped[str] = mapped_column(ForeignKey("course_content.id"), primary_key=True)

    module_id: Mapped[str] = mapped_column(ForeignKey("modules.id"))
    module: Mapped["Module"] = relationship(back_populates="pages")
    position: Mapped[int]

    page_type: Mapped[ModulePageType]

    __mapper_args__ = {"polymorphic_identity": "page"}


class Module(Base):
    __tablename__ = "modules"

    id: Mapped[str] = mapped_column(primary_key=True)

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="modules")

    name: Mapped[str]
    pages: Mapped[List["ModulePage"]] = relationship(
        back_populates="module",
        order_by=ModulePage.position,
        cascade="save-update, merge, delete",
    )
    position: Mapped[int]

    @property
    def content_items(self) -> Sequence[CourseContentItem]:
        """
        Alias for .pages
        """
        return self.pages


class Course(Base):
    __tablename__ = "courses"

    id: Mapped[str] = mapped_column(primary_key=True)

    name: Mapped[str]
    configuration: Mapped["CourseConfiguration"] = relationship(
        back_populates="course", cascade="save-update, merge, delete"
    )
    resources: Mapped[List["Resource"]] = relationship(
        back_populates="course",
        order_by=Resource.discovery_date,
        cascade="save-update, merge, delete",
    )
    modules: Mapped[List["Module"]] = relationship(
        back_populates="course",
        order_by=Module.position,
        cascade="save-update, merge, delete",
    )
    assignment_groups: Mapped[List["AssignmentGroup"]] = relationship(
        back_populates="course",
        order_by=AssignmentGroup.position,
        cascade="save-update, merge, delete",
    )
    content_items: Mapped[List["CourseContentItem"]] = relationship(
        back_populates="course",
        order_by=CourseContentItem.creation_date,
        cascade="save-update, merge, delete",
    )
    messages: Mapped[List["CourseMessage"]] = relationship(
        viewonly=True,
        primaryjoin=id == CourseMessage.course_id,
        order_by=CourseMessage.creation_date,
    )
    assignments: Mapped[List["Assignment"]] = relationship(
        viewonly=True,
        primaryjoin=id == Assignment.course_id,
        order_by=Assignment.due_date,
    )
    panopto_folder_id: Mapped[Optional[str]]
    term_id: Mapped[str] = mapped_column(ForeignKey("terms.id"))
    term: Mapped["Term"] = relationship(back_populates="courses")


class Term(Base):
    __tablename__ = "terms"

    id: Mapped[str] = mapped_column(primary_key=True)

    start_date: Mapped[Optional[datetime]]
    end_date: Mapped[Optional[datetime]]
    name: Mapped[str]
    courses: Mapped[List["Course"]] = relationship(
        back_populates="term",
        order_by=Course.name,
        cascade="save-update, merge, delete",
    )

    def __hash__(self):
        return (
            hash(self.id)
            ^ hash(self.start_date)
            ^ hash(self.end_date)
            ^ hash(self.name)
        )



from .canvas_data_sanitiser import remove_unwanted_whitespaces
from .collecting_task_group import CollectingTaskGroup
from .flatten import flatten
from .form_input_extractor import extract_form_inputs
from .tzremover import strip_tz



from dataclasses import dataclass

import qcanvas_backend.database.types as db


@dataclass
class PageResources:
    page_id: str
    resources: list[db.Resource]
    invisible_resources: list[db.Resource]



import asyncio
from pathlib import Path
from typing import AsyncIterator, Callable

import asynchronize
import yt_dlp
from httpx import URL
from qcanvas_api_clients.panopto import PanoptoClient
from yt_dlp import DownloadError

from qcanvas_backend.net.resources.download.download_progress import DownloadProgress


async def download_video(
    video_id: str, client: PanoptoClient, destination: Path
) -> AsyncIterator[DownloadProgress]:

    max_retries = 3
    retries = 0

    while retries < max_retries:
        retries += 1

        callback = asynchronize.AsyncCallback()
        download_task = asyncio.create_task(
            asyncio.to_thread(
                _download_helper,
                headers={"cookie": cookies_for_download(client)},
                url=client.get_viewer_url(video_id),
                destination=destination,
                step_callback=callback.step_callback,
                done_callback=callback.finished_callback,
            )
        )

        yield DownloadProgress(0, 0)
        finished = False

        async for message in callback:
            if _is_progress_message(message):
                progress = _unwrap_args(message)

                current = int(progress["downloaded_bytes"])
                total = int(progress["total_bytes"])

                if current != total or not finished:
                    if current == total:
                        finished = True

                    yield DownloadProgress(
                        downloaded_bytes=current,
                        total_bytes=total,
                    )
        try:
            await download_task

            # Ensure we indicate we are finished
            if not finished:
                yield DownloadProgress(downloaded_bytes=1, total_bytes=1)

            return
        except DownloadError as e:
            if _is_logged_out_error(e):
                await client.force_authenticate()
                continue
            else:
                raise RuntimeError() from e
        except Exception as e:
            raise RuntimeError() from e

    if retries == max_retries:
        raise RuntimeError(
            f"Gave up trying to download panopto video after {max_retries} attempts"
        )


def _is_progress_message(message: asynchronize.Args) -> bool:
    if len(message.args) >= 1 and isinstance(message.args[0], dict):
        data = _unwrap_args(message)

        return "total_bytes" in data and "downloaded_bytes" in data

    return False


def _unwrap_args(args: asynchronize.Args) -> dict:
    return args.args[0]


def _is_logged_out_error(e: DownloadError):
    return "This video is only available for registered users" in e.msg


def _download_helper(
    url: URL,
    headers: dict[str, str],
    destination: Path,
    done_callback: Callable,
    step_callback: Callable,
):
    try:
        with yt_dlp.YoutubeDL(
            dict(
                progress_hooks=[step_callback],
                http_headers=headers,
                outtmpl=str(destination.absolute()),
            )
        ) as ytdl:
            ytdl.download(str(url))
    finally:
        done_callback()


def cookies_for_download(client: PanoptoClient) -> str:
    return ";".join(
        [f"{cookie.name}={cookie.value}" for cookie in client.cookies_for_download]
    )



from bs4 import Tag


def extract_form_inputs(form_element: Tag) -> dict[str, str]:
    result = {}

    for element in form_element.find_all("input"):
        if _is_form_input_valid(element):
            result[element.attrs["id"]] = element.attrs["value"]

    return result


def _is_form_input_valid(element: Tag) -> bool:
    return "id" in element.attrs and "value" in element.attrs



import logging
from typing import Any

import qcanvas_api_clients.canvas as canvas
from sqlalchemy.ext.asyncio import AsyncSession

import qcanvas_backend.database.types as db
import qcanvas_backend.gql_queries as gql
from qcanvas_backend.database.types import ModulePageType
from qcanvas_backend.net.canvas import CourseMailItem, PageWithContent
from qcanvas_backend.net.sync._internal.canvas_sync_observer import CanvasSyncObservable
from qcanvas_backend.util import remove_unwanted_whitespaces, strip_tz

_logger = logging.getLogger(__name__)


class APIDataImporter(CanvasSyncObservable):
    """
    An APIDataImporter accepts canvas API objects and converts and adds them to the database
    """

    _type_map = {
        gql.Course: db.Course,
        gql.Term: db.Term,
        gql.Module: db.Module,
        gql.Assignment: db.Assignment,
        gql.AssignmentGroup: db.AssignmentGroup,
        gql.File: db.ModulePage,
        PageWithContent: db.ModulePage,
        CourseMailItem: db.CourseMessage,
        canvas.Announcement: db.CourseMessage,
    }

    @staticmethod
    def _id_from_api_object(obj: Any) -> str | int:
        if isinstance(obj, (canvas.Announcement, CourseMailItem)):
            return obj.id
        elif hasattr(obj, "q_id"):
            return obj.q_id
        else:
            raise TypeError(f"{type(obj)} can not be handled")

    def __init__(self, session: AsyncSession):
        super().__init__()
        self._session = session

    async def convert_and_store_course(
        self, course: gql.Course, panopto_folder_id: str | None
    ):
        db_course: db.Course = await self._find_or_create_db_entry(course)
        is_new = self._is_record_new(db_course)

        db_course.id = course.q_id
        db_course.name = remove_unwanted_whitespaces(course.name) or "No title"
        db_course.term_id = course.term.q_id

        if panopto_folder_id is not None:
            db_course.panopto_folder_id = panopto_folder_id

        if is_new:
            db_course.configuration = db.CourseConfiguration()
            self.notify_observers_for_updated_item(db_course)

    async def convert_and_store_module(
        self, module: gql.Module, course_id: str, position: int
    ):
        db_module: db.Module = await self._find_or_create_db_entry(module)

        is_new = self._is_record_new(db_module)

        db_module.id = module.q_id
        db_module.name = module.name or "No title"
        db_module.course_id = course_id
        db_module.position = position

        if is_new:
            self.notify_observers_for_updated_item(db_module)

    async def convert_and_store_page(self, page: PageWithContent):
        db_page: db.ModulePage = await self._find_or_create_db_entry(page)

        if not (self._is_record_new_or_updated(db_object=db_page, api_object=page)):
            # Updating the record is not going to do anything
            return

        db_page.id = page.q_id
        db_page.course_id = page.course.q_id
        db_page.module_id = page.module.q_id
        db_page.name = remove_unwanted_whitespaces(page.name) or "No title"
        db_page.body = remove_unwanted_whitespaces(page.content)
        db_page.creation_date = page.created_at
        db_page.last_modification_date = page.updated_at
        db_page.position = page.position
        db_page.page_type = ModulePageType.PAGE

        self.notify_observers_for_updated_item(db_page)

    async def convert_and_store_term(self, term: gql.Term):
        db_term: db.Term = await self._find_or_create_db_entry(term)

        is_new = self._is_record_new(db_term)

        db_term.id = term.q_id
        db_term.end_date = term.end_at
        db_term.start_date = term.start_at
        db_term.name = remove_unwanted_whitespaces(term.name) or "No title"

        if is_new:
            self.notify_observers_for_updated_item(db_term)

    async def convert_and_store_assignment_group(
        self, assignment_group: gql.AssignmentGroup, course_id: str, position: int
    ):
        db_group: db.AssignmentGroup = await self._find_or_create_db_entry(
            assignment_group
        )

        is_new = self._is_record_new(db_group)

        db_group.id = assignment_group.q_id
        db_group.name = remove_unwanted_whitespaces(assignment_group.name) or "No title"
        db_group.course_id = course_id
        db_group.group_weight = assignment_group.group_weight
        db_group.position = position

        if is_new:
            self.notify_observers_for_updated_item(db_group)

    async def convert_and_store_assignment(
        self, assignment: gql.Assignment, group_id: str
    ):
        db_assignment: db.Assignment = await self._find_or_create_db_entry(assignment)

        # todo the assignment submission object also has an updated date now, I'm not sure if we need to check both
        #  that, as well as the assignment's modify date?
        if not (
            self._is_record_new_or_updated(
                db_object=db_assignment, api_object=assignment
            )
        ):
            # Updating the record is not going to do anything
            return

        db_assignment.id = assignment.q_id
        db_assignment.course_id = assignment.course_id
        db_assignment.name = remove_unwanted_whitespaces(assignment.name) or "No title"
        db_assignment.body = remove_unwanted_whitespaces(assignment.description)
        db_assignment.creation_date = assignment.created_at
        db_assignment.last_modification_date = assignment.updated_at
        db_assignment.due_date = assignment.due_at
        db_assignment.position = assignment.position
        db_assignment.group_id = group_id
        db_assignment.max_mark_possible = assignment.points_possible
        db_assignment.mark = self._get_assignment_mark(assignment)

        self.notify_observers_for_updated_item(db_assignment)

    @staticmethod
    def _get_assignment_mark(assignment: gql.Assignment) -> float | None:
        if assignment.submissions_connection is not None:
            submissions = assignment.submissions_connection.nodes
            if len(submissions) > 1:
                _logger.warning(
                    "Assignment %s has multiple submissions, expected only 1",
                    assignment.q_id,
                )

            if len(submissions) >= 1:
                return submissions[0].score

        return None

    async def convert_and_store_announcement(self, announcement: canvas.Announcement):
        db_message: db.CourseMessage = await self._find_or_create_db_entry(announcement)

        # Canvas doesn't seem to track last modification date on announcements for some reason
        # Otherwise, I would check for modification time here
        is_new = self._is_record_new(db_message)

        db_message.id = str(announcement.id)
        db_message.course_id = announcement.course_id
        db_message.creation_date = announcement.created_at
        db_message.name = remove_unwanted_whitespaces(announcement.title) or "No title"
        db_message.body = remove_unwanted_whitespaces(announcement.message)
        db_message.sender_name = announcement.user_name
        db_message.has_been_read = False

        if is_new:
            self.notify_observers_for_updated_item(db_message)

    async def convert_and_store_mail_item(self, mail: CourseMailItem):
        db_message: db.CourseMessage = await self._find_or_create_db_entry(mail)

        is_new = self._is_record_new(db_message)

        db_message.id = mail.id
        db_message.course_id = mail.course_id
        db_message.creation_date = mail.date
        db_message.name = remove_unwanted_whitespaces(mail.subject) or "No title"
        db_message.body = self._convert_plaintext_to_html(
            remove_unwanted_whitespaces(mail.body)
        )
        db_message.sender_name = mail.author_name
        db_message.has_been_read = False

        if is_new:
            self.notify_observers_for_updated_item(db_message)

    def _convert_plaintext_to_html(self, plaintext: str) -> str:
        return plaintext.replace("\n", "<br/>\n")

    async def _find_or_create_db_entry(self, obj: Any):
        obj_type = type(obj)

        if obj_type not in self._type_map.keys():
            raise TypeError(f"{obj_type} is not present in _type_map")

        obj_id = self._id_from_api_object(obj)
        db_type = self._type_map[obj_type]
        db_object = await self._session.get(db_type, obj_id)

        if db_object is None:
            _logger.debug(
                'Converting new %s (api) to %s (db) (id="%s")',
                obj_type.__name__,
                db_type.__name__,
                obj_id,
            )
            db_object = db_type()
            self._session.add(db_object)
        else:
            _logger.debug(
                'Found existing %s (db) for %s (api) (id="%s")',
                db_type.__name__,
                obj_type.__name__,
                obj_id,
            )

        return db_object

    @staticmethod
    def _is_record_new_or_updated(
        db_object: Any | db.ModificationDate, api_object: Any
    ):
        if APIDataImporter._is_record_new(db_object):
            return True
        elif hasattr(api_object, "updated_at") and isinstance(
            db_object, db.ModificationDate
        ):
            return db_object.last_modification_date < strip_tz(api_object.updated_at)
        else:
            return False

    @staticmethod
    def _is_record_new(db_obj: Any):
        return db_obj.id is None
